

#include "snet.h"
#include "support.h"


/****************************************************************************
 * 
 * SNET_InitServer - This initializes a server.  It does not actually start 
 * the server (Call SNET_StartServer for that).  We can not start the server 
 * yet because the data packet registry is not ready yet.

 */

SNet_Client * SNET_InitClient (char* server_name, int port)
  {
  int loop;
  SNet_Client *client;

  // Need to make sure that SDL is initialized!
  if (SDL_Init(0) == -1)
    {
    printf("SDL_Init: %s\n", SDL_GetError());
    return NULL;
    }
  if (SDLNet_Init() == -1)
    {
    printf("SDLNet_Init: %s\n", SDLNet_GetError());
    return NULL;
    }



  client = (SNet_Client*)malloc(sizeof(SNet_Client));


  if (client == NULL)
    return NULL;


  // Store the port for later
  client->port = port;
  
  // Fill in connection information so we can attempt to open the port later!
  if (SDLNet_ResolveHost(&client->ip,server_name, client->port) == -1)
    {
    // The server name is invalid (Does not exist) or the port is bad
    free (client);
    return NULL;
    }

  // Need to init the StringQs 
  client->str_in.head = client->str_in.tail = NULL;
  client->str_out.head = client->str_out.tail = NULL;
  
  // Need to init the CommandQs 
  client->cmd_in.head = client->cmd_in.tail = NULL;
  client->cmd_out.head = client->cmd_out.tail = NULL;

  // Need to init the DataQ
  client->dataq.head = client->dataq.tail = NULL;
 


  // Need to init Mutexs for strQ and cmdQ
  client->str_in.hold = SDL_CreateMutex();
  client->str_out.hold = SDL_CreateMutex();
  client->cmd_in.hold = SDL_CreateMutex();
  client->cmd_out.hold = SDL_CreateMutex();
  client->dataq.hold = SDL_CreateMutex();

  // Init the data handlers
  for (loop = 0; loop < 65536; loop++)
    {
    client->handler[loop].pack = NULL;
    client->handler[loop].unpack = NULL;
    client->handler[loop].num = -1;  // Not in use!
    client->handler[loop].len = -1;
    client->handler[loop].blocker = NULL;
    }

  return client;
  }








/* This function starts the client in a separate thread and then returns to 
   the caller so the main program can continue. */
int SNET_StartClient(SNet_Client *client)
  {
  SDL_Thread * process;
 
  client->tcpsock = SDLNet_TCP_Open(&client->ip);

  if (!client->tcpsock)
    {
    // Open Failed!
    return -2;
    }


  
  // Spawn child process
  process = SDL_CreateThread(SNET_RunClient, client);
  if (process == NULL)
    {
    // Not sure if this can happen, no documentation for thread creation 
    // failure!
    return -3;
    }


  return 0;
  }











/* This function is called after the client is attached to the port.  It 
 * handles the main event loop for the client 
 */
int SNET_RunClient(void *c)
  {
  
  int ready;
  
  int loop;

  Uint8 type;         // Holds the type of packet we read in from the port

  Uint16 cmd;         // Used for I/O with cmd packets
  Uint32 cmddata;

  Uint16 len;         // Used for I/O with String packets
 
  Uint16 which;        // Used for I/O with Data Packets

  unsigned char *buffer;       // Used for String packets and Data packets

  Uint16 who;            // Temp for which client we are accessing

  SNet_Client *client;

  client = (SNet_Client*)c;

  loop = 0;
  buffer = NULL;
  
  client->set = SDLNet_AllocSocketSet(1);
  SDLNet_TCP_AddSocket(client->set, client->tcpsock);
  // Shouldn't be able to get an error from this, it only holds one, and we are
  // giving it only 1!

  while (1)
    {
    // Need to see if there is any input from the socket
    ready = SDLNet_CheckSockets(client->set, 0);
    if (ready == -1)
      {
      printf("SDLNet_CheckSockets: %s\n", SDLNet_GetError());
      // most of the time this is a system error, where perror might help you.
      perror("SDLNet_CheckSockets");
      // An error occured.  Not sure what to do about it!
      }
    if (ready > 0)
      {
      if (SDLNet_SocketReady(client->tcpsock))
	{
	if (SNET_ClientRead8(client, &type) == -1) 
	  {
	  return -1;
	  }
	
	switch (type)
	  {
	  case 1:             // Command packet
	    if (SNET_ClientRead16(client, &cmd) == -1) return -1;
	    if (SNET_ClientRead32(client, &cmddata) == -1) return -1;
	    // Add this command to the incoming CmdQ
	    SNET_CmdEnqueue(&client->cmd_in, cmd, cmddata, loop);
	    break;
	  case 2:             // String packet
	    if (SNET_ClientRead16(client, &len) == -1) return -1;
	    buffer = (unsigned char*)malloc(len + 1);
	    SNET_ClientReadN(client, buffer, len);
	    buffer[len] = '\0';  // Add a NULL terminator
	    SNET_StringEnqueue(&client->str_in, buffer);
	    free(buffer);
	    break;
	  case 3:             // Data packet
	    if (SNET_ClientRead16(client, &which) == -1) return -1;
	    buffer = (unsigned char*)malloc(client->handler[which].len);
	    if (SNET_ClientReadN(client,buffer,client->handler[which].len)!=-1)
	      {
	      SDL_LockMutex(client->handler[which].blocker);
	      client->handler[which].unpack(buffer);
	      SDL_UnlockMutex(client->handler[which].blocker);
	      }
	    else
	      {
	      free(buffer);
	      return -1;
	      }
	    free(buffer);
	    break;
	  }
	}
      } // If there was input from the socket
      
	



    // Need to send any data out to the server
    while (SNET_DataDequeue(&client->dataq, &who, &which) == 0)
      {
      buffer = (unsigned char*)malloc(client->handler[which].len);
      SDL_LockMutex(client->handler[which].blocker);
      client->handler[which].pack(buffer);
      SDL_UnlockMutex(client->handler[which].blocker);
      // Send the information out!
      if (SNET_SendData(client->tcpsock, which, buffer, 
		       client->handler[which].len) == 0)
	{
	SDLNet_TCP_Close(client->tcpsock);
	
	// Need to notify the program about a lost connection
	SNET_CmdEnqueue(&client->cmd_in, 0,5,0);
	free(buffer);
	return -1;
	}
      free(buffer);
      }  // End while Data packets


    // Need to send any pending strings out to the server
    while ((buffer= SNET_StringDequeue(&client->str_out, &len)) != NULL)
      {
      if (SNET_SendString(client->tcpsock, buffer, len)== 0)
	{
	// The send failed.  Remove this socket!
	SDLNet_TCP_Close(client->tcpsock);
	// Need to notify the program about a lost connection
	SNET_CmdEnqueue(&client->cmd_in, 0,6,0);
	return -1;
	}
	  	
      free(buffer);
      } // End while strings to send!


    // Need to send any pending commands out to server
    while (SNET_CmdDequeue(&client->cmd_out, &cmd, &cmddata, &which) != -1)
      {
      // Send this command out to client
      if (SNET_SendCommand(client->tcpsock, cmd, cmddata) == 0)
	{
	// The send failed.  Remove this socket!
	SDLNet_TCP_Close(client->tcpsock);
	// Need to notify the program about a lost connection
	SNET_CmdEnqueue(&client->cmd_in, 0,7,0);
	return -1;
	}
      }


    } // End while


  // Can't get here!
  return -1;

  } // End client function 










// These two functions are user level functions for getting and receiving 
// Commands from the Client.


/* This function will poll the clients incoming command queue and get the next
   command to be processed (if there is one) back from the client.  The 
   function is non-blocking, so it is safe to call as part of an event loop.
   CMD is the command send, DATA is the 32 data associated with the command. */
   
int SNET_PollClient(SNet_Client *client,Uint16 *cmd, Uint32* data)
  {
  Uint16 who; // For a client, Who is always 0.
  return SNET_CmdDequeue(&client->cmd_in, cmd, data, &who);
  }



/* This function sends a command from the client application to the server 
   CMD is the command 
   number to send.  DATA is 32 bits of data associated with the command. */
int SNET_SendCmdToServer(SNet_Client *client, Uint16 cmd, Uint32 data)
  {
  return SNET_CmdEnqueue (&client->cmd_out, cmd, data, 0);
  }









/* This function is used by the client application to send a string to all the
   connected clients (via server). */
int SNET_ClientSendString(SNet_Client *client, unsigned char* buffer)
  {
  return SNET_StringEnqueue(&client->str_out, buffer);
  }





unsigned char* SNET_ClientGetString(SNet_Client *client)
  {
  Uint16 len;
  
 

  return SNET_StringDequeue(&client->str_in, &len);
 
  }
	






/* Register a data-handler with the client.  Each data handler has a unique 
   identifier (0-65535).  The application needs to supply two functions one to 
   pack the data, and one to unpack the data to and from the network.  The 
   user must also specify the len of the NETWORK DATA PACKET.  This is most 
   likely NOT the same as sizeof (datatype).  We must do this to avoid issues 
   of endian and differences in data alignment in structures across different
   platforms.  Len can be computed by summing the memory required for each 
   part of the structure. Actually, as a parting thought, this is more 
   efficient (network bandwidth wise) than sending the whole structure, as we
   don't have to send the padding! */

int SNET_ClientRegisterHandler(SNet_Client *client, Uint16 which,
			 void (*pack)(unsigned char* buffer),
			 void (*unpack)(unsigned char* buffer), 
			 int len, SDL_mutex* blocker)
  {
  if (pack == NULL) return -2; // Pack function does not exist!
  if (unpack == NULL) return -3; // Unpack function does not exist!
  if (blocker == NULL) return -4; // Mutex does not exist!

  client->handler[which].num = which;
  client->handler[which].pack = pack;
  client->handler[which].unpack = unpack;
  client->handler[which].len = len;
  client->handler[which].blocker = blocker;

  return 0;
  }



/* Request that the client send data out to the server.  This function 
   is called by the client application to initiate transfer of a data structure
   across the network to the server.  Note:  This function does not 
   pack the data, it only enqueues the request for the client to deal with it
   later. 
*/
int SNET_SendDataToServer(SNet_Client *client, int which)
  {
  return SNET_DataEnqueue(&client->dataq, 0, which);
  }
