

#include "snet.h"
#include "support.h"






/****************************************************************************
 * 
 * SNET_InitServer - This initializes a server.  It does not actually start 
 * the server (Call SNET_StartServer for that).  We can not start the server 
 * yet because the data packet registry is not ready yet.

 */

SNet_Server * SNET_InitServer( int port, int maxclients)
  {
  int loop;
  SNet_Server *server;

  // Need to make sure that SDL is initialized!
  if (SDL_Init(0) == -1)
    {
    printf("SDL_Init: %s\n", SDL_GetError());
    return NULL;
    }
  if (SDLNet_Init() == -1)
    {
    printf("SDLNet_Init: %s\n", SDLNet_GetError());
    return NULL;
    }
  
  server = (SNet_Server*)malloc(sizeof(SNet_Server));

   if (server == NULL)
    return NULL;
  
  // Store the port for later
  server->port = port;
  
  // So that we can deny extra connection attempts!
  server->maxclients = maxclients + 1;  // Extra one is for server socket
 
  // Need to init the StringQs 
  server->strq.head = server->strq.tail = NULL;
 
  // Need to init the CommandQs 
  server->cmd_in.head = server->cmd_in.tail = NULL;
  server->cmd_out.head = server->cmd_out.head = NULL;
  
  // Need to init the DataQ
  server->dataq.head = server->dataq.tail = NULL;

  // Need to init Mutexs for strQ and cmdQ
  server->strq.hold = SDL_CreateMutex();
  server->cmd_in.hold = SDL_CreateMutex();
  server->cmd_out.hold = SDL_CreateMutex();
  server->dataq.hold = SDL_CreateMutex();

  // Init the data handlers
  for (loop = 0; loop < 65536; loop++)
    {
    server->handler[loop].pack = NULL;
    server->handler[loop].unpack = NULL;
    server->handler[loop].num = -1;  // Not in use!
    server->handler[loop].len = -1;
    server->handler[loop].blocker = NULL;
    }

  // Set up the socketset and the list of clients
  server->set = SDLNet_AllocSocketSet(server->maxclients);
  for (loop = 0; loop < server->maxclients; loop++)
    {
    server->sockets[loop].active = 0;
    }
	
   return server;
  }








/* This function starts the server in a separate thread and then returns to 
   the caller so the main program can continue. */
int SNET_StartServer(SNet_Server *server)
  {
  SDL_Thread * process;



  // Attach server to port get ready to accept connections
  
    if (SDLNet_ResolveHost(&server->ip,NULL, server->port) == -1)
    {
    // ResolveHost Failed (shouldn't happen... I would think...
    printf("Uh Oh\n");
    return -1;
    }
  server->tcpsock = SDLNet_TCP_Open(&server->ip);
  if (!server->tcpsock)
    {
    // Open Failed!
    printf("Server open failed!\n");
    return -2;
    }

  SDLNet_TCP_AddSocket(server->set, server->tcpsock);


  
  // Spawn child process
  process = SDL_CreateThread(SNET_RunServer, server);
  if (process == NULL)
    {
    // Not sure if this can happen, no documentation for thread creation 
    // failure!
    return -3;
    }

  return 0;
  }











/* This function is called after the server is attached to the port.  It 
 * handles the main event loop for the server 
 */
int SNET_RunServer(void *s)
  {
  
  TCPsocket client;
  int ready;
 
  int loop;

  Uint8 type;         // Holds the type of packet we read in from the port

  Uint16 cmd;         // Used for I/O with cmd packets
  Uint32 cmddata;

  Uint16 len;         // Used for I/O with String packets
 
  Uint16 which;        // Used for I/O with Data Packets

  unsigned char *buffer;       // Used for String packets and Data packets

  Uint16 who;            // Temp for which client we are accessing

  SNet_Server *server;

  server = (SNet_Server*)s;


  buffer = NULL;
  while (1)
    {
    
    // Check all the sockets to see if there are incoming packets or errors 
    // on them.
    ready = SDLNet_CheckSockets(server->set, 0);
    if (ready == -1)
      {
      printf("SDLNet_CheckSockets: %s\n", SDLNet_GetError());
      // most of the time this is a system error, where perror might help you.
      perror("SDLNet_CheckSockets");
      // An error occured.  Not sure what to do about it!
      }
    else if (ready >0 )
      {
      // Check all the sockets in the set to see if they are active
      for (loop = 0; loop < server->maxclients; loop++)
	{
	if ((server->sockets[loop].active == 1) &&
	    SDLNet_SocketReady(server->sockets[loop].sock))
	  {
	  // This socket has some activity, so get type
	  if (SNET_ServerRead8(server,loop,&type)!=-1)//Handles failure in func
	    {
	    switch (type)
	      {
	      case 1:             // Command packet
		if (SNET_ServerRead16(server, loop, &cmd) == -1) break;
		if (SNET_ServerRead32(server, loop, &cmddata) == -1) break;
		// Add this command to the incoming CmdQ
		SNET_CmdEnqueue(&server->cmd_in, cmd, cmddata, loop);
		break;
	      case 2:             // String packet
		len = -17;
		if (SNET_ServerRead16(server, loop, &len) == -1) break;
		buffer = (unsigned char*)malloc(len + 1);
		if (SNET_ServerReadN(server, loop, buffer, len) == -1) break;
		buffer[len] = '\0';  // Add a NULL terminator
		SNET_StringEnqueue(&server->strq, buffer);
		free(buffer);
		break;
	      case 3:             // Data packet
		if (SNET_ServerRead16(server,loop,&which) == -1) break;
		buffer = (unsigned char*)malloc(server->handler[which].len);
		if (SNET_ServerReadN(server,loop,buffer,
				     server->handler[which].len) != -1)
		  {
		  SDL_LockMutex(server->handler[which].blocker);
		  server->handler[which].unpack(buffer);
		  SDL_UnlockMutex(server->handler[which].blocker);
		  }
		free(buffer);
		break;
	      }
	    }
	  }
	} // End For each client in list




      // Check to see if we have a new connection
      if (SDLNet_SocketReady(server->tcpsock))
	{

	client = SDLNet_TCP_Accept(server->tcpsock);
	if (client)
	  {
	  // Need to add this client to the socketset
	  if (SDLNet_TCP_AddSocket(server->set, client) > -1)
	    {
	    // This socket was added successfully
	    // Need to add this client to the list of socket list
	    for (loop = 0; loop < server->maxclients; loop++)
	      {
	      if (server->sockets[loop].active == 0)
		{
		server->sockets[loop].sock = client;

		server->sockets[loop].active = 1;
		// Need to notify the main program about the new client!
		SNET_CmdEnqueue(&server->cmd_in, 1, loop, loop);
                loop = server->maxclients + 10;
		}
	      }
	    }
	  else
	    {
            // Too many connections.  Should send the child a note, and 
            // close the socket!
	    SNET_SendCommand(client, 0, 0);
	    SDLNet_TCP_Close(client);
	    }
	  }
	} // End add new client socket
      } // End CheckSockets


    // Need to send any data out to the sockets!
    while (SNET_DataDequeue(&server->dataq, &who, &which) == 0)
      {
      if (server->sockets[who].active == 1)
	{
	buffer = (unsigned char*)malloc(server->handler[which].len);
	
	SDL_LockMutex(server->handler[which].blocker);
	server->handler[which].pack(buffer);
	SDL_UnlockMutex(server->handler[which].blocker);
	
	// Send the information out!
	if (SNET_SendData(server->sockets[who].sock, which, buffer, 
			  server->handler[which].len) == 0)
	  {
	  SDLNet_TCP_DelSocket(server->set, server->sockets[who].sock);
	  SDLNet_TCP_Close(server->sockets[who].sock);
	  server->sockets[who].active = 0;
			  
	  // Need to notify the program about a lost client
	  SNET_CmdEnqueue(&server->cmd_in, 0, who, who);    
	  }
	}  // End while Data packets
      }

    // Need to send any pending strings out to the sockets
    while ((buffer=SNET_StringDequeue(&server->strq, &len)) != NULL)
      {
      for (loop = 0; loop < server->maxclients; loop++)
	{
	if (server->sockets[loop].active == 1)
	  {
	  if (SNET_SendString(server->sockets[loop].sock, buffer, len) == 0 )
	    {
	    // The send failed.  Remove this socket!
	    SDLNet_TCP_DelSocket(server->set, server->sockets[who].sock);
	    SDLNet_TCP_Close(server->sockets[who].sock);
	    server->sockets[who].active = 0;
	    
	    // Need to notify the program about a lost client
	    SNET_CmdEnqueue(&server->cmd_in, 0, who, who);
	    }
	  }
	}
      free(buffer);
      } // End while strings to send!


    // Need to send any pending commands out to sockets
    while (SNET_CmdDequeue(&server->cmd_out, &cmd, &cmddata, &who) != -1)
      {
      if (server->sockets[who].active == 1)
	{
	// Send this command out to client
	if (SNET_SendCommand(server->sockets[who].sock, cmd, cmddata) == 0)
	  {
	  // The send failed.  Remove this socket!
	  SDLNet_TCP_DelSocket(server->set, server->sockets[who].sock);
	  SDLNet_TCP_Close(server->sockets[who].sock);
	  server->sockets[who].active = 0;
	  
	  // Need to notify the program about a lost client
	  SNET_CmdEnqueue(&server->cmd_in, 0, who, who);
	  }
	}
      }


    } // End while


  // Can't get here!
  return -1;

  } // End server function 










// These two functions are user level functions for getting and receiving 
// Commands from the Server.


/* This function will poll the servers incoming command queue and get the next
   command to be processed (if there is one) back from the server.  The 
   function is non-blocking, so it is safe to call as part of an event loop.
   CMD is the command send, DATA is the 32 data associated with the command, 
   and WHO is the client number (this tells us who the command came from). */
int SNET_PollServer(SNet_Server *server,Uint16 *cmd, Uint32* data, Uint16 *who)
  {
  return  SNET_CmdDequeue(&server->cmd_in, cmd, data, who);
  }



/* This function sends a command from the server application to a specific 
   client.  WHO specifies the client to get the command.  CMD is the command 
   number to send.  DATA is 32 bits of data associated with the command. */
int SNET_SendCmdToClient(SNet_Server *server, Uint16 who, Uint16 cmd, 
			 Uint32 data)
  {
  return SNET_CmdEnqueue (&server->cmd_out, cmd, data, who);
  }









/* This function is used by the server application to send a string to all the
   connected clients. */
int SNET_ServerSendString(SNet_Server *server, unsigned char* buffer)
  {
  return SNET_StringEnqueue(&server->strq, buffer);
  }









/* Register a data-handler with the server.  Each data handler has a unique 
   identifier (0-65535).  The application needs to supply two functions one to 
   pack the data, and one to unpack the data to and from the network.  The 
   user must also specify the len of the NETWORK DATA PACKET.  This is most 
   likely NOT the same as sizeof (datatype).  We must do this to avoid issues 
   of endian and differences in data alignment in structures across different
   platforms.  Len can be computed by summing the memory required for each 
   part of the structure. Actually, as a parting thought, this is more 
   efficient (network bandwidth wise) than sending the whole structure, as we
   don't have to send the padding! */

int SNET_ServerRegisterHandler(SNet_Server *server, Uint16 which,
			 void (*pack)(unsigned char* buffer),
			 void (*unpack)(unsigned char* buffer), 
			 int len, SDL_mutex* blocker)
  {
  if (pack == NULL) return -2; // Pack function does not exist!
  if (unpack == NULL) return -3; // Unpack function does not exist!
  if (blocker == NULL) return -4; // Mutex does not exist!

  server->handler[which].num = which;
  server->handler[which].pack = pack;
  server->handler[which].unpack = unpack;
  server->handler[which].len = len;
  server->handler[which].blocker = blocker;

  return 0;
  }











/* Request that the server send data out to a specific client.  This function 
   is called by the server application to initiate transfer of a data structure
   across the network to a specific client.  Note:  This function does not 
   pack the data, it only enqueues the request for the server to deal with it
   later. 
*/
int SNET_SendDataToClient(SNet_Server *server, int who, int which)
  {
  return SNET_DataEnqueue(&server->dataq, who, which);
  }






// *************************************************************************
// * 
// * PACKING AND UNPACKING FUNCTIONS
// * 
// * Provided as a support to users.  Use of these functions to pack and 
// * unpack data is optional.  These routines use a big-endian packing order
// * 
// *************************************************************************


void SNET_pack_int(unsigned char* buffer, int data)
  {
  buffer[0] = (unsigned char)((data & 0xff000000) >> 24);
  buffer[1] = (unsigned char)((data & 0x00ff0000) >> 16);
  buffer[2] = (unsigned char)((data & 0x0000ff00) >> 8);
  buffer[3] = (unsigned char)((data & 0x000000ff));
  }

int SNET_unpack_int(unsigned char* buffer)
  {
  return  (buffer[0] * 16777216) + (buffer[1] * 65536) + (buffer[2] * 256)
    + buffer[3];
  }
	 

void SNET_pack_short(unsigned char* buffer, short data)
  {
  buffer[0] = (unsigned char) ((data & 0xff00) >> 8);
  buffer[1] = (unsigned char) ((data & 0x00ff));
  }

short SNET_unpack_short(unsigned char* buffer)
  {
  return (buffer[0] * 256) + buffer[1];
  }
