#include "SDL.h"
#include "SDL_thread.h"
#include "SDL_net.h"

#include "stdlib.h"




/* We want to support C and C++, so tell compiler not to mangle these function
 * names!  */

#ifdef __cplusplus
extern "C" {
#endif




/* A Data_Registry is a structure that holds information for dealing with a 
   Type 3 packed (Preformated data).  The structure holds pointers to packing
   and unpacking functions, as well as the size of the data bufffer transmitted
   (likely not the same as sizeof(struct), and a Mutex to ensure reads and 
   writes do not overlap. */ 
struct Data_Registry
  {
  int  num;                              // Which data item this is
  
  void (*pack) (unsigned char* buffer);  // Pointer to function to pack data
                                         // data buffer with the data
  
  void (*unpack) (unsigned char* buffer);// Pointer to function to unpack
                                         // the data from the buffer and put
                                         // it into the structure
  
  int len;                               // Len of the data IN the buffer
                                         // This is probably NOT sizeof the
                                         // the data structure.
  
  SDL_mutex* blocker;                    // We must insist that ANY access to
                                         // the core data be handled via Mutex
                                         // Note: The mutex only solves 
                                         // access problems on this machines, 
                                         // it does nothing to solve the 
                                         // problem across the network
  };  
typedef struct Data_Registry data_registry;







/* A String_Entry is a substructure for the stringQ which holds incoming and 
   outgoing strings. */
struct String_Entry
  {
  unsigned char* buffer;                 // Actual data for the string.  Should
                                         // be NULL terminated
  
  Uint16 len;                             // Len of the string in the buffer 
                                         // including the NULL character
  };
typedef struct String_Entry string_entry;
  
  
  
struct StringNode
  {
  string_entry data;
  struct StringNode* next;
  };
  
typedef struct StringNode SNET_StringNode;



/* The String_Q holds the incoming Strings from external sources or a local
   client.  The main loop for the server will handle pulling them from the Q
   and sending them to the clients.  This Q acts as a buffer between the 
   network threads and the main program.  */
struct String_Q
  {
  SNET_StringNode* head;
  SNET_StringNode* tail;
  SDL_mutex* hold;
  };
typedef struct String_Q SNet_StringQ;



struct DataNode
  {
  Uint16 what;
  Uint16 who;
  struct DataNode* next;
  };
typedef struct DataNode SNET_DataNode;


/* A Queue to hold outgoing data packet entries.  The user specifys the which
   data handler to use, and who is going to get the packet.  The server takes 
   care of actually handling the sending.
*/
struct Data_Q
  {
  SNET_DataNode* head;
  SNET_DataNode* tail;
  SDL_mutex* hold;
  };
typedef struct Data_Q SNet_DataQ;

  

struct CmdNode
  {
  Uint16 cmd;
  Uint32 data;
  Uint16 who;
  struct CmdNode* next;
  };
typedef struct CmdNode SNET_CmdNode;

/* The Command_Q holds a Queue of commands from external sources or from the 
   local client.  The main loop for the server will extract them from the Q
   and handle them appropriately. */
struct Command_Q
  {
  SNET_CmdNode* head;
  SNET_CmdNode* tail;
  SDL_mutex* hold;
  };
typedef struct Command_Q SNet_CommandQ;



#define MAX_CLIENT 100

struct SNET_socket_registry
  {
  TCPsocket sock;
  char active;                   // Whether this socket is in use
  };
typedef struct SNET_socket_registry SNet_SocketRegistry;







struct SNET_server
  {  
  SNet_StringQ strq;                  // Outgoing strings
  SNet_DataQ dataq;                   // Outgoing data packets
  
  SNet_CommandQ cmd_in;               // Incoming commands
  SNet_CommandQ cmd_out;              // Outgoing commands
  
  
  
  data_registry handler[65536];      // Hold 1 short worth of items!
  
  int port;                           // Port this thing is attached to
  
  int maxclients;

  IPaddress ip;
  TCPsocket tcpsock;                  // Our tcp socket
  
  // Need an AVL tree or something to hold all the tcpsockets of our 
  // Clients
  SNet_SocketRegistry sockets[MAX_CLIENT]; // All the sockets that are "live"
  
  
  SDLNet_SocketSet set;               // Socket set for checking active socks
  
  }; 
typedef struct SNET_server SNet_Server;




struct SNET_client
  {
  SNet_StringQ str_in;
  SNet_StringQ str_out;
  SNet_CommandQ cmd_in;
  SNet_CommandQ cmd_out;

  SNet_DataQ dataq;                   // Outgoing data packets
  
  data_registry handler[65536];       // Hold 1 short worth of items!
  
  int port;                           // Port this thing is attached to
  IPaddress ip;
  TCPsocket tcpsock;                  // Our tcp socket
  SDLNet_SocketSet set;
  };
typedef struct SNET_client SNet_Client;








SNet_Server*  SNET_InitServer( int port, int maxclients);

int SNET_StartServer(SNet_Server *server);

int SNET_RunServer(void *server);

int SNET_PollServer(SNet_Server *server,Uint16 *cmd,Uint32* data,Uint16 *who);

int SNET_SendCmdToClient(SNet_Server *server,Uint16 who,Uint16 cmd,
			 Uint32 data);


//int SNET_SendString(TCPsocket sock, unsigned char* buffer, int len);



int SNET_ServerRegisterHandler(SNet_Server *server, Uint16 which,
			 void (*pack)(unsigned char* buffer),
			 void (*unpack)(unsigned char* buffer), 
			 int len, SDL_mutex* blocker);



int SNET_SendDataToClient(SNet_Server *server, int who, int which);

int SNET_ServerSendString(SNet_Server *server, unsigned char* buffer);





SNet_Client* SNET_InitClient(char* server_name, int port);

int SNET_StartClient(SNet_Client *client);

int SNET_RunClient(void *c);

int SNET_PollClient(SNet_Client *client,Uint16 *cmd, Uint32* data);

int SNET_SendCmdToServer(SNet_Client *client, Uint16 cmd, Uint32 data);

int SNET_ClientSendString(SNet_Client *client, unsigned char* buffer);

unsigned char* SNET_ClientGetString(SNet_Client *client);

int SNET_ClientRegisterHandler(SNet_Client *client, Uint16 which,
			       void (*pack)(unsigned char* buffer),
			       void (*unpack)(unsigned char* buffer), 
			       int len, SDL_mutex* blocker);

int SNET_SendDataToServer(SNet_Client *client, int which);
 
void SNET_pack_int(unsigned char* buffer, int data);
int SNET_unpack_int(unsigned char* buffer);
void SNET_pack_short(unsigned char* buffer, short data);
short SNET_unpack_short(unsigned char* buffer);



/* Need to end the brace for C++ unmangling support */
#ifdef __cplusplus
}
#endif

