#include "snet.h"

#include <string.h>
#include "support.h"

/* Read 1 byte from the network socket.  If there is an error, close the 
   socket and tell the application */
int SNET_ServerRead8(SNet_Server *server, Uint16 which, unsigned char *data)
  {
  int result;
  unsigned char info;

  result = SDLNet_TCP_Recv(server->sockets[which].sock, &info, 1);
  if (result < 1)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_DelSocket(server->set, server->sockets[which].sock);
    SDLNet_TCP_Close(server->sockets[which].sock);
    server->sockets[which].active = 0;
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&server->cmd_in, 0, which, which);
    return -1;
    }

  *data = info;
  return 1;
  }






/* Read a 2-byte value from the network socket.  Convert it to local endian.
   If there is an error, close the socket and tell the application. */
int SNET_ServerRead16(SNet_Server *server, Uint16 which, Uint16 *data)
  {
  int result;
  unsigned char info[5];

  result = SDLNet_TCP_Recv(server->sockets[which].sock, info, 2);
  if (result < 2)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_DelSocket(server->set, server->sockets[which].sock);
    SDLNet_TCP_Close(server->sockets[which].sock);
    server->sockets[which].active = 0;
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&server->cmd_in, 0, which, which);
    return -1;
    }

  *data = SDLNet_Read16(info);
  return 1;
  }





/* Read a 4-byte value from the network socket.  Convert it to local endian.
   If there is a an error, close the socket and tell the application. */
int SNET_ServerRead32(SNet_Server *server, Uint16 which, Uint32 *data)
  {
  int result;
  unsigned char info[5];

  result = SDLNet_TCP_Recv(server->sockets[which].sock, info, 4);
  if (result < 4)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_DelSocket(server->set, server->sockets[which].sock);
    SDLNet_TCP_Close(server->sockets[which].sock);
    server->sockets[which].active = 0;
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&server->cmd_in, 0, which, which);
    return -1;
    }

  *data = SDLNet_Read32(info);
  return 1;
  }



/* Read a n-byte value from the network socket.  Data is assumed to be a 
   sequence of chars... IE no endian conversion is done.  The function is 
   used to read an entire String at once.  It will also be used to read the
   data for a data packet (although it will not convert the data-packet to 
   the data structure). */
int SNET_ServerReadN(SNet_Server *server, Uint16 which,
	unsigned char* buffer,Uint16 len)
  {
  int result;
  
  result = SDLNet_TCP_Recv(server->sockets[which].sock, buffer, len);

  if (result < len)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_DelSocket(server->set, server->sockets[which].sock);
    SDLNet_TCP_Close(server->sockets[which].sock);
    server->sockets[which].active = 0;
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&server->cmd_in, 0, which, which);
    return -1;
    }

  return 1;
  }










/* Send a command out a socket.  Function returns 1 if successful, 0 if 
   failure (probably disconnected socket). */

int SNET_SendCommand(TCPsocket sock, Uint16 cmd, Uint32 data)
  {
  unsigned char buffer[10];

  buffer[0] = 1; // This is a command being sent out!
  SDLNet_Write16(cmd, &buffer[1]);
  SDLNet_Write32(data, &buffer[3]);

  return SDLNet_TCP_Send(sock, buffer, 7) == 7;
  }



/* Send a string packet out a socket.  Function returns len if successfull, 
   0 if failure (probably a disconnect). */
int SNET_SendString(TCPsocket sock, unsigned char* buffer, Uint16 len)
  {
  unsigned char type[10];
  type[0] = 2;  // String packet

  int tmp;
  SDLNet_Write16(len, &type[1]);  // String len
  // Send type of packet

  tmp = SDLNet_TCP_Send(sock, type, 3);
  tmp = SDLNet_TCP_Send(sock, buffer, len);
  return (tmp == len);
  }


int SNET_SendData(TCPsocket sock, Uint16 which, unsigned char* buffer, int len)
  {
  unsigned char temp[10];  // Packet header

  temp[0] = 3;  // This is a data packet
  SDLNet_Write16(which, &temp[1]);
  

  if (SDLNet_TCP_Send(sock, temp, 3) != 3) return 0;

  return  SDLNet_TCP_Send(sock, buffer, len) == len;
  }











/* Read 1 byte from the network socket.  If there is an error, close the 
   socket and tell the application */
int SNET_ClientRead8(SNet_Client *client, unsigned char *data)
  {
  int result;
  unsigned char info;

  result = SDLNet_TCP_Recv(client->tcpsock, &info, 1);
  if (result < 1)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_Close(client->tcpsock);

    // Need to notify client about a lost connection!
    SNET_CmdEnqueue(&client->cmd_in, 0,1,0);
    return -1;
    }

  *data = info;
  return 1;
  }






/* Read a 2-byte value from the network socket.  Convert it to local endian.
   If there is an error, close the socket and tell the application. */
int SNET_ClientRead16(SNet_Client *client, Uint16 *data)
  {
  int result;
  unsigned char info[5];

  result = SDLNet_TCP_Recv(client->tcpsock, info, 2);
  if (result < 2)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_Close(client->tcpsock);
    
    // Need to notify client about a lost connection!
    SNET_CmdEnqueue(&client->cmd_in, 0,2,0);
    return -1;
    }

  *data = SDLNet_Read16(info);
  return 1;
  }





/* Read a 4-byte value from the network socket.  Convert it to local endian.
   If there is a an error, close the socket and tell the application. */
int SNET_ClientRead32(SNet_Client *client, Uint32 *data)
  {
  int result;
  unsigned char info[5];

  result = SDLNet_TCP_Recv(client->tcpsock, info, 4);
  if (result < 4)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_Close(client->tcpsock);
   
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&client->cmd_in, 0,3,0);
    return -1;
    }

  *data = SDLNet_Read32(info);
  return 1;
  }



/* Read a n-byte value from the network socket.  Data is assumed to be a 
   sequence of chars... IE no endian conversion is done.  The function is 
   used to read an entire String at once.  It will also be used to read the
   data for a data packet (although it will not convert the data-packet to 
   the data structure). */
int SNET_ClientReadN(SNet_Client *client, unsigned char* buffer, Uint16 len)
  {
  int result;
  
  result = SDLNet_TCP_Recv(client->tcpsock, buffer, len);
  if (result < len)
    {
    // Socket had a read error.  We lost the connection.  Kill it.
    SDLNet_TCP_Close(client->tcpsock);
  
    // Need to notify the program about a lost client
    SNET_CmdEnqueue(&client->cmd_in, 0,4,0);
    return -1;
    }

  return 1;
  }



// *************************************************************************
// *
// * QUEUE functions here!!
// * 
// *************************************************************************


/* Take a command and add it to a CmdQ.  Commands coming in from clients sit 
   here until the main application gets to them.  Commands to clients sit here
   until the server event loop gets to sending them out.  

   LL implementation so we don't worry about filling up.
*/

int SNET_CmdEnqueue(SNet_CommandQ *q, Uint16 cmd, Uint32 data, Uint16 who)
  {
  SNET_CmdNode* temp;

  temp= (SNET_CmdNode*)malloc(sizeof(SNET_CmdNode));

  temp->next = NULL;
  temp->cmd = cmd;
  temp->data = data;
  temp->who = who;

  SDL_LockMutex(q->hold);

  if (q->head == NULL)
    {
    q->head = q->tail = temp;
    SDL_UnlockMutex(q->hold);
    return 0;
    }

  q->tail->next = temp;
  q->tail = temp;

  SDL_UnlockMutex(q->hold);
  return 0;
  }



/* The function is used to remove a pending command from a CmdQ.  The server
   application makes use of this by calling SNET_PollServer.  The server event
   loop does a while !empty on this to send all pending commands out to 
   clients.
   
   LL implementation so we don't worry about filling up.
*/
int SNET_CmdDequeue(SNet_CommandQ *q, Uint16 *cmd, Uint32 *data, Uint16 *who)
  {
  SNET_CmdNode* temp;

  SDL_LockMutex(q->hold);

  if (q->head == NULL)
    {
    // Q is empty
    SDL_UnlockMutex(q->hold);
    return -1;
    }
  
  *cmd = q->head->cmd;
  *data = q->head->data;
  *who = q->head->who;


  temp = q->head;
  if (q->head == q->tail)
    {
    q->head = q->tail = NULL;
    }
  else
    {
    q->head = q->head->next;
    }
  free(temp);

  SDL_UnlockMutex(q->hold);

  return 0;
  }




/* Take a string and add it to a StringQ.  

   An LL implementation saves worrying about filling the Q!
*/

int SNET_StringEnqueue(SNet_StringQ *q, unsigned char* string)
  {
  SNET_StringNode* temp;
  SDL_LockMutex(q->hold);
  temp = (SNET_StringNode*)malloc(sizeof(SNET_StringNode));
  temp->next = NULL;
  temp->data.len = (Uint16)(strlen(string)+1);
  temp->data.buffer = (unsigned char*)malloc(temp->data.len);
  strcpy(temp->data.buffer, string);
  

  
  if (q->head == NULL)
    {
    q->head = q->tail = temp;
    SDL_UnlockMutex(q->hold);
    return 0;
    }

  q->tail->next = temp;
  q->tail = temp;

  SDL_UnlockMutex(q->hold);
   
  return 0;
  }




/* The function is used to obtain a string from the StringQ.  The function 
   assumes that STRING is a pointer to a char, and not an array.  The function
   will allocate memory to hold the copy.
   
  Now an LL implementation so we don't have to worry about overflow!
*/
char* SNET_StringDequeue(SNet_StringQ *q, Uint16 *len)
  {
  SNET_StringNode * temp;
  char* string;
  SDL_LockMutex(q->hold);
 
  if (q->head == NULL)
    {
    // Q us empty
    SDL_UnlockMutex(q->hold);
    *len = 0;
    return NULL;
    }
 
  string = (unsigned char*)malloc(q->head->data.len);
  strcpy(string, q->head->data.buffer);

  *len = q->head->data.len;

  free(q->head->data.buffer);
  temp = q->head;


  if (q->head == q->tail)
    {
    // Last thing in Q
    q->head = q->tail = NULL;
    }
  else
    {
    q->head = q->head->next;
    }

  free(temp);

  SDL_UnlockMutex(q->hold);
 
  return string;
  }




/* Take a data packet output request and add it to a DataQ.  

   Now an LL implementation so we don't have to worry about overflow.
*/

int SNET_DataEnqueue(SNet_DataQ *q, Uint16 who, Uint16 what )
  {
  SNET_DataNode* temp;
  temp = (SNET_DataNode*)malloc(sizeof(SNET_DataNode));
  temp->next = NULL;
  temp->what = what;
  temp->who = who;

  SDL_LockMutex(q->hold);

  if (q->head == NULL)
    {
    q->head = q->tail = temp;
    SDL_UnlockMutex(q->hold);
    return 0;
    }

  q->tail->next = temp;
  q->tail = temp;

  SDL_UnlockMutex(q->hold);
   
  return 0;
  }




/* The function is used to obtain information on a data packet to send out. 
   The who is which client to send it to, and what is which data handler to 
   use.  Note, the data has not been packed prior to this call!
   
   Now an LL implementation so that we don't have to worry about overflow
*/
int SNET_DataDequeue(SNet_DataQ *q, Uint16 *who, Uint16 *what)
  {
  SNET_DataNode* temp;

  SDL_LockMutex(q->hold);

  if (q->head == NULL)
    {
    // Q is empty
    SDL_UnlockMutex(q->hold);
    return -1;
    }

  *who = q->head->who;
  *what = q->head->what;

  temp = q->head;
  
  if (q->head == q->tail)
    {
    q->head = q->tail = NULL;
    }
  else
    {
    q->head = q->head->next;
    }
  free(temp);
  
  SDL_UnlockMutex(q->hold);

  return 0;
  }






